@@ -36,7 +36,6 @@ struct _PixbufDocument
 
 	GdkPixbuf *pixbuf;
 	GdkDrawable *target;
-	gdouble scale;
 
 	gint x_offset, y_offset;
 };
@@ -93,92 +92,31 @@ pixbuf_document_get_n_pages (EvDocument  *document)
 	return 1;
 }
 
-static void
-pixbuf_document_set_page (EvDocument  *document,
-			  int          page)
-{
-	/* Do nothing */
-}
-
-static int
-pixbuf_document_get_page (EvDocument  *document)
-{
-	return 1;
-}
-
-static void
-pixbuf_document_set_target (EvDocument  *document,
-			    GdkDrawable *target)
-{
-	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
-
-	pixbuf_document->target = target;
-}
-
-static void
-pixbuf_document_set_scale (EvDocument  *document,
-			   double       scale)
-{
-	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
-
-	pixbuf_document->scale = scale;
-}
-
-static void
-pixbuf_document_set_page_offset (EvDocument  *document,
-			      int          x,
-			      int          y)
-{
-	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
-
-	pixbuf_document->x_offset = x;
-	pixbuf_document->y_offset = y;
-}
-
 static void
 pixbuf_document_get_page_size (EvDocument   *document,
 			       int           page,
-			       int          *width,
-			       int          *height)
+			       double       *width,
+			       double       *height)
 {
 	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
 
 	if (width)
-		*width = gdk_pixbuf_get_width (pixbuf_document->pixbuf) * pixbuf_document->scale;
+		*width = gdk_pixbuf_get_width (pixbuf_document->pixbuf);
 	if (height)
-		*height = gdk_pixbuf_get_height (pixbuf_document->pixbuf) * pixbuf_document->scale;
+		*height = gdk_pixbuf_get_height (pixbuf_document->pixbuf);
+
+	printf ("get_page_size, page=%d, *width=%f, *height=%f\n",
+		page, *width, *height);
 }
 
-static void
-pixbuf_document_render (EvDocument  *document,
-			int          clip_x,
-			int          clip_y,
-			int          clip_width,
-			int          clip_height)
+static GdkPixbuf*
+pixbuf_document_render_pixbuf (EvDocument  *document, int page, double scale)
 {
 	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
-	GdkPixbuf *tmp_pixbuf;
-	
-	tmp_pixbuf = gdk_pixbuf_new (gdk_pixbuf_get_colorspace (pixbuf_document->pixbuf),
-				     gdk_pixbuf_get_has_alpha (pixbuf_document->pixbuf),
-				     gdk_pixbuf_get_bits_per_sample (pixbuf_document->pixbuf),
-				     clip_width, clip_height);
-	
-	gdk_pixbuf_fill (tmp_pixbuf, 0xffffffff);
-	gdk_pixbuf_scale (pixbuf_document->pixbuf, tmp_pixbuf, 0, 0,
-			  MIN(gdk_pixbuf_get_width(pixbuf_document->pixbuf)* pixbuf_document->scale-clip_x, clip_width),
-			  MIN(gdk_pixbuf_get_height(pixbuf_document->pixbuf)* pixbuf_document->scale-clip_y, clip_height),
-			  -clip_x, -clip_y,
-			  pixbuf_document->scale, pixbuf_document->scale,
-			  GDK_INTERP_BILINEAR);
-	
-	gdk_draw_pixbuf (pixbuf_document->target, NULL, tmp_pixbuf,
-			 0, 0,
-			 clip_x, clip_y,
-			 clip_width, clip_height, GDK_RGB_DITHER_NORMAL,
-			 0, 0);
-
-	g_object_unref (tmp_pixbuf);
+	return gdk_pixbuf_scale_simple (pixbuf_document->pixbuf,
+					gdk_pixbuf_get_width (pixbuf_document->pixbuf) * scale,
+					gdk_pixbuf_get_height (pixbuf_document->pixbuf) * scale,
+					GDK_INTERP_BILINEAR);
 }
 
 static void
@@ -232,54 +170,40 @@ pixbuf_document_class_init (PixbufDocumentClass *klass)
 }
 
 static char *
-pixbuf_document_get_text (EvDocument *document, GdkRectangle *rect)
+pixbuf_document_get_text (EvDocument *document, int page, EvRectangle *rect)
 {
 	/* FIXME this method should not be in EvDocument */
 	g_warning ("pixbuf_document_get_text not implemented");
 	return NULL;
 }
 
-
-static EvLink *
-pixbuf_document_get_link (EvDocument *document,
-		          int         x,
-		          int	      y)
-{
-	return NULL;
-}
-
 static void
 pixbuf_document_document_iface_init (EvDocumentIface *iface)
 {
 	iface->load = pixbuf_document_load;
 	iface->save = pixbuf_document_save;
 	iface->get_text = pixbuf_document_get_text;
-	iface->get_link = pixbuf_document_get_link;
 	iface->get_n_pages = pixbuf_document_get_n_pages;
-	iface->set_page = pixbuf_document_set_page;
-	iface->get_page = pixbuf_document_get_page;
-	iface->set_scale = pixbuf_document_set_scale;
-	iface->set_target = pixbuf_document_set_target;
-	iface->set_page_offset = pixbuf_document_set_page_offset;
 	iface->get_page_size = pixbuf_document_get_page_size;
-	iface->render = pixbuf_document_render;
+	iface->render_pixbuf = pixbuf_document_render_pixbuf;
 }
 
 static GdkPixbuf *
 pixbuf_document_thumbnails_get_thumbnail (EvDocumentThumbnails   *document,
-					  gint 			 page,
-					  gint			 width)
+					  gint 			  page,
+					  gint			  size,
+					  gboolean                border)
 {
 	PixbufDocument *pixbuf_document = PIXBUF_DOCUMENT (document);
 	GdkPixbuf *pixbuf;
 	gdouble scale_factor;
 	gint height;
 	
-	scale_factor = (gdouble)width / gdk_pixbuf_get_width (pixbuf_document->pixbuf);
+	scale_factor = (gdouble)size / gdk_pixbuf_get_width (pixbuf_document->pixbuf);
 
 	height = gdk_pixbuf_get_height (pixbuf_document->pixbuf) * scale_factor;
 	
-	pixbuf = gdk_pixbuf_scale_simple (pixbuf_document->pixbuf, width, height,
+	pixbuf = gdk_pixbuf_scale_simple (pixbuf_document->pixbuf, size, height,
 					  GDK_INTERP_BILINEAR);
 	
 	return pixbuf;
@@ -312,8 +236,6 @@ pixbuf_document_document_thumbnails_iface_init (EvDocumentThumbnailsIface *iface
 static void
 pixbuf_document_init (PixbufDocument *pixbuf_document)
 {
-	pixbuf_document->scale = 1.0;
-
 	pixbuf_document->x_offset = 0;
 	pixbuf_document->y_offset = 0;
 }
