@@ -32,6 +32,7 @@
 #include "ev-sidebar-thumbnails.h"
 #include "ev-document-thumbnails.h"
 #include "ev-document-misc.h"
+#include "ev-job-queue.h"
 #include "ev-window.h"
 #include "ev-utils.h"
 
@@ -46,15 +47,14 @@ struct _EvSidebarThumbnailsPrivate {
 	GtkListStore *list_store;
 	EvDocument *document;
 
-	guint idle_id;
-	gint current_page, n_pages, pages_done;
-	GtkTreeIter current_page_iter;
+	gint n_pages, pages_done;
 };
 
 enum {
 	COLUMN_PAGE_STRING,
 	COLUMN_PIXBUF,
 	COLUMN_THUMBNAIL_SET,
+	COLUMN_JOB,
 	NUM_COLUMNS
 };
 
@@ -63,21 +63,6 @@ G_DEFINE_TYPE (EvSidebarThumbnails, ev_sidebar_thumbnails, GTK_TYPE_VBOX);
 #define EV_SIDEBAR_THUMBNAILS_GET_PRIVATE(object) \
 	(G_TYPE_INSTANCE_GET_PRIVATE ((object), EV_TYPE_SIDEBAR_THUMBNAILS, EvSidebarThumbnailsPrivate));
 
-static void
-ev_sidebar_thumbnails_destroy (GtkObject *object)
-{
-	EvSidebarThumbnails *ev_sidebar_thumbnails = EV_SIDEBAR_THUMBNAILS (object);
-	EvSidebarThumbnailsPrivate *priv = ev_sidebar_thumbnails->priv;
-
-	if (priv->idle_id != 0) {
-		g_source_remove (priv->idle_id);
-
-		priv->idle_id = 0;
-	}
-
-	GTK_OBJECT_CLASS (ev_sidebar_thumbnails_parent_class)->destroy (object);
-}
-
 static void
 ev_sidebar_thumbnails_class_init (EvSidebarThumbnailsClass *ev_sidebar_thumbnails_class)
 {
@@ -87,10 +72,7 @@ ev_sidebar_thumbnails_class_init (EvSidebarThumbnailsClass *ev_sidebar_thumbnail
 	g_object_class = G_OBJECT_CLASS (ev_sidebar_thumbnails_class);
 	gtk_object_class = GTK_OBJECT_CLASS (ev_sidebar_thumbnails_class);
 
-	gtk_object_class->destroy = ev_sidebar_thumbnails_destroy;
-
 	g_type_class_add_private (g_object_class, sizeof (EvSidebarThumbnailsPrivate));
-
 }
 
 static void
@@ -112,18 +94,11 @@ adjustment_changed_cb (GtkAdjustment       *adjustment,
 		return;
 
 	page = gtk_tree_path_get_indices (path)[0];
-	if (page == priv->current_page)
-		return;
 	gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->list_store),
 				 &iter, path);
 	gtk_tree_model_get (GTK_TREE_MODEL (priv->list_store), &iter,
 			    COLUMN_THUMBNAIL_SET, &thumbnail_set,
 			    -1);
-	if (! thumbnail_set) {
-		priv->current_page = page;
-		priv->current_page_iter = iter;
-		
-	}
 }
 
 static void
@@ -131,28 +106,23 @@ ev_sidebar_tree_selection_changed (GtkTreeSelection *selection,
 				   EvSidebarThumbnails *ev_sidebar_thumbnails)
 {
 	EvSidebarThumbnailsPrivate *priv;
-	GtkWidget *window;
+	EvPageCache *page_cache;
 	GtkTreePath *path;
 	GtkTreeIter iter;
 	int page;
 
 	priv = ev_sidebar_thumbnails->priv = EV_SIDEBAR_THUMBNAILS_GET_PRIVATE (ev_sidebar_thumbnails);
-  
+
 	if (!gtk_tree_selection_get_selected (selection, NULL, &iter))
 		return;
-	
+
 	path = gtk_tree_model_get_path (GTK_TREE_MODEL (priv->list_store),
 					&iter);
-
-	page = gtk_tree_path_get_indices (path)[0] + 1;
-
+	page = gtk_tree_path_get_indices (path)[0];
 	gtk_tree_path_free (path);
 
-	window = gtk_widget_get_ancestor (GTK_WIDGET (ev_sidebar_thumbnails),
-					  EV_TYPE_WINDOW);
-	if (window && ev_document_get_page (priv->document) != page) {
-		ev_window_open_page (EV_WINDOW (window), page);
-	}
+	page_cache = ev_document_get_page_cache (priv->document);
+	ev_page_cache_set_current_page (page_cache, page);
 }
 
 static void
@@ -164,10 +134,10 @@ ev_sidebar_thumbnails_init (EvSidebarThumbnails *ev_sidebar_thumbnails)
 	GtkTreeSelection *selection;
 
 	priv = ev_sidebar_thumbnails->priv = EV_SIDEBAR_THUMBNAILS_GET_PRIVATE (ev_sidebar_thumbnails);
-	
-	priv->list_store = gtk_list_store_new (NUM_COLUMNS, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_BOOLEAN);
+
+	priv->list_store = gtk_list_store_new (NUM_COLUMNS, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_BOOLEAN, EV_TYPE_JOB_THUMBNAIL);
 	priv->tree_view = gtk_tree_view_new_with_model (GTK_TREE_MODEL (priv->list_store));
-	
+
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->tree_view));
 	g_signal_connect (selection, "changed",
 			  G_CALLBACK (ev_sidebar_tree_selection_changed), ev_sidebar_thumbnails);
@@ -211,126 +181,39 @@ ev_sidebar_thumbnails_new (void)
 	return ev_sidebar_thumbnails;
 }
 
-static gboolean
-do_one_iteration (EvSidebarThumbnails *ev_sidebar_thumbnails)
-{
-	EvSidebarThumbnailsPrivate *priv = ev_sidebar_thumbnails->priv;
-	GdkPixbuf *pixbuf;
-	gboolean thumbnail_set;
-
-	gtk_tree_model_get (GTK_TREE_MODEL (priv->list_store),
-			    &(priv->current_page_iter),
-			    COLUMN_THUMBNAIL_SET, &thumbnail_set,
-			    -1);
-	if (!thumbnail_set) {
-		pixbuf = ev_document_thumbnails_get_thumbnail
-				(EV_DOCUMENT_THUMBNAILS (priv->document),
-				 priv->current_page, THUMBNAIL_WIDTH, TRUE);
-
-		gtk_list_store_set (priv->list_store,
-				    &(priv->current_page_iter),
-				    COLUMN_PIXBUF, pixbuf,
-				    COLUMN_THUMBNAIL_SET, TRUE,
-				    -1);
-
-		g_object_unref (pixbuf);
-		priv->pages_done ++;
-	}
-
-	priv->current_page++;
-
-	if (priv->current_page == priv->n_pages) {
-		priv->current_page = 0;
-		gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->list_store),
-					       &(priv->current_page_iter));
-	} else {
-		gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->list_store),
-					  &(priv->current_page_iter));
-	}
-
-	if (priv->pages_done == priv->n_pages)
-		return FALSE;
-	else
-		return TRUE;
-}
-
-static gboolean
-populate_thumbnails_idle (gpointer data)
-{
-	GTimer *timer;
-	int i;
-	gdouble time_elapsed = 0;
-
-	EvSidebarThumbnails *ev_sidebar_thumbnails = EV_SIDEBAR_THUMBNAILS (data);
-	EvSidebarThumbnailsPrivate *priv = ev_sidebar_thumbnails->priv;
-
-#if PROFILE_THUMB == 1
-	static GTimer *total_timer;
-	static gboolean first_time = TRUE;
-
-	if (first_time) {
-		total_timer = g_timer_new ();
-		first_time = FALSE;
-		g_timer_start (total_timer);
-	}
-#endif
-
-	/* undo the thumbnailing idle and handler */
-	if (priv->pages_done == priv->n_pages) {
-		priv->idle_id = 0;
-		g_signal_handlers_disconnect_by_func (priv->vadjustment,
-						      adjustment_changed_cb,
-						      ev_sidebar_thumbnails);
-#if PROFILE_THUMB == 1
-		time_elapsed = g_timer_elapsed (total_timer, NULL);
-		g_timer_destroy (total_timer);
-		g_print ("%d rows done in %f seconds\n",
-			 gtk_tree_model_iter_n_children (GTK_TREE_MODEL (priv->list_store), NULL),
-			 time_elapsed);
-#endif
-		return FALSE;
-	}
-
-	timer = g_timer_new ();
-	i = 0;
-	g_timer_start (timer);
-	while (do_one_iteration (ev_sidebar_thumbnails)) {
-		i++;
-		time_elapsed = g_timer_elapsed (timer, NULL);
-		if (time_elapsed > IDLE_WORK_LENGTH/1000000)
-			break;
-	}
-	g_timer_destroy (timer);
-#if PROFILE_THUMB == 2
-	g_print ("%d rows done this idle in %f seconds\n", i, time_elapsed);
-#endif
-
-	return TRUE;
-}
-
-void
-ev_sidebar_thumbnails_select_page (EvSidebarThumbnails *sidebar,
-				   int                  page)
+static void
+page_changed_cb (EvPageCache         *page_cache,
+		 int                  page,
+		 EvSidebarThumbnails *sidebar)
 {
 	GtkTreePath *path;
 	GtkTreeSelection *selection;
 
-	/* if the EvSidebar's document can't provide thumbnails */
-	if (sidebar->priv->document == NULL) 
-		return;
-
-	path = gtk_tree_path_new_from_indices (page - 1, -1);
+	path = gtk_tree_path_new_from_indices (page, -1);
 	selection = gtk_tree_view_get_selection
 			(GTK_TREE_VIEW (sidebar->priv->tree_view));
 
-	if (path) {
-		gtk_tree_selection_select_path (selection, path);
-		gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (sidebar->priv->tree_view),
-					      path, NULL, FALSE, 0.0, 0.0);
-		gtk_tree_path_free (path);	
-	}
+	gtk_tree_selection_select_path (selection, path);
+	gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (sidebar->priv->tree_view),
+				      path, NULL, FALSE, 0.0, 0.0);
+	gtk_tree_path_free (path);
 }
 
+static void
+thumbnail_job_completed_callback (EvJobThumbnail      *job,
+				  EvSidebarThumbnails *sidebar_thumbnails)
+{
+	EvSidebarThumbnailsPrivate *priv = sidebar_thumbnails->priv;
+	GtkTreeIter *iter;
+
+	iter = (GtkTreeIter *) g_object_get_data (G_OBJECT (job), "tree_iter");
+	gtk_list_store_set (priv->list_store,
+			    iter,
+			    COLUMN_PIXBUF, job->thumbnail,
+			    COLUMN_THUMBNAIL_SET, TRUE,
+			    COLUMN_JOB, NULL,
+			    -1);
+}
 
 void
 ev_sidebar_thumbnails_set_document (EvSidebarThumbnails *sidebar_thumbnails,
@@ -339,46 +222,62 @@ ev_sidebar_thumbnails_set_document (EvSidebarThumbnails *sidebar_thumbnails,
 	GdkPixbuf *loading_icon;
 	gint i, n_pages;
 	GtkTreeIter iter;
-	gchar *page;
 	gint width = THUMBNAIL_WIDTH;
 	gint height = THUMBNAIL_WIDTH;
+	EvPageCache *page_cache;
 
 	EvSidebarThumbnailsPrivate *priv = sidebar_thumbnails->priv;
 
 	g_return_if_fail (EV_IS_DOCUMENT_THUMBNAILS (document));
 
-	if (priv->idle_id != 0) {
-		g_source_remove (priv->idle_id);
-	}
-	n_pages = ev_document_get_n_pages (document);
+	page_cache = ev_document_get_page_cache (document);
+	n_pages = ev_page_cache_get_n_pages (page_cache);
 
 	priv->document = document;
-	priv->idle_id = g_idle_add (populate_thumbnails_idle, sidebar_thumbnails);
 	priv->n_pages = n_pages;
 
 	/* We get the dimensions of the first doc so that we can make a blank
 	 * icon.  */
+	g_mutex_lock (EV_DOC_MUTEX);
 	ev_document_thumbnails_get_dimensions (EV_DOCUMENT_THUMBNAILS (priv->document),
 					       1, THUMBNAIL_WIDTH, &width, &height);
+	g_mutex_unlock (EV_DOC_MUTEX);
+
 	loading_icon = ev_document_misc_get_thumbnail_frame (width, height, NULL);
 
 	gtk_list_store_clear (priv->list_store);
+	for (i = 0; i < n_pages; i++) {
+		EvJob *job;
+		gchar *page_label;
+		gchar *page_string;
+
+		job = ev_job_thumbnail_new (priv->document, i, THUMBNAIL_WIDTH);
+		page_label = ev_page_cache_get_page_label (page_cache, i);
+		page_string = g_markup_printf_escaped ("<i>%s</i>", page_label);
 
-	for (i = 1; i <= n_pages; i++) {
-		page = g_strdup_printf ("<i>%d</i>", i);
 		gtk_list_store_append (priv->list_store, &iter);
 		gtk_list_store_set (priv->list_store, &iter,
-				    COLUMN_PAGE_STRING, page,
+				    COLUMN_PAGE_STRING, page_string,
 				    COLUMN_PIXBUF, loading_icon,
 				    COLUMN_THUMBNAIL_SET, FALSE,
+				    COLUMN_JOB, job,
 				    -1);
-		g_free (page);
+		g_free (page_label);
+		g_free (page_string);
+		ev_job_queue_add_job (job, EV_JOB_PRIORITY_LOW);
+		g_object_set_data_full (G_OBJECT (job), "tree_iter",
+					gtk_tree_iter_copy (&iter),
+					(GDestroyNotify) gtk_tree_iter_free);
+		g_signal_connect (job, "finished",
+				  G_CALLBACK (thumbnail_job_completed_callback),
+				  sidebar_thumbnails);
 	}
 
 	g_object_unref (loading_icon);
-	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->list_store),
-				       &(priv->current_page_iter));
-	priv->current_page = 0;
-	priv->pages_done = 0;
+
+	/* Connect to the signal and trigger a fake callback */
+	g_signal_connect (page_cache, "page-changed", G_CALLBACK (page_changed_cb), sidebar_thumbnails);
+	page_changed_cb (page_cache, ev_page_cache_get_current_page (page_cache), sidebar_thumbnails);
+
 }
 
